from typing import Union

from vivarium.core.process import Process
from vivarium.core.types import Schema, State, Update
from vivarium.composites.toys import ToyProcess




def make_logging_process(
        process_class,
        logging_port_name="log_update"
) -> type:
    """
    Given a subclass of Process, returns a new subclass that behaves exactly
    the same except that it logs all of its updates in a port with name given by
    logging_port_name.

    The returned class has the same name as process_class, but prefixed with
    'Logging_'.

    Args:
        process_class: The Process class to be logged
        logging_port_name: Name of the port in which updates will be stored
                           ('log_update' by default.)

    Returns:
        logging_process: the logging version of process_class.
    """

    if not issubclass(process_class, Process):
        raise ValueError(f'process_class must be a subclass of Process.')

    logging_process = type(f"Logging_{process_class.__name__}",
                           (process_class,),
                           {})
    __class__ = logging_process  # set __class__ manually so super() knows what to do

    def ports_schema(
            self
    ) -> Schema:
        ports = super().ports_schema()  # type: ignore
        ports[logging_port_name] = {'_default': {}, '_updater': 'set', '_emit': True}  # add a new port
        return ports

    def next_update(
            self,
            timestep: Union[float, int],
            states: State
    ) -> Update:
        update = super().next_update(timestep, states)  # type: ignore
        log_update = {logging_port_name: update}  # log the update
        return {**update, **log_update}

    logging_process.ports_schema = ports_schema  # type: ignore
    logging_process.next_update = next_update  # type: ignore

    return logging_process


def test_logging_process():
    logging_toy = make_logging_process(ToyProcess)
    logging_toy_instance = logging_toy()

    ports = logging_toy_instance.ports_schema()
    assert 'log_update' in ports


if __name__ == '__main__':
    test_logging_process()



from vivarium.core.engine import Engine, pf
from vivarium.core.control import run_library_cli
from vivarium.experiments.engine_tests import get_toy_transport_in_env_composite
from vivarium.core.process import Process, Step
from vivarium.core.types import (
    Schema, Union, Update, State, Flow, Topology)
from vivarium.processes.meta_division import daughter_phylogeny_id


class TopView(Step):

    defaults = {
        'division_threshold': 6.0
    }

    def ports_schema(self) -> Schema:
        return {
            'top': '**',
            'other': {
                '_default': 2.0
            }
        }

    def next_update(
            self, timestep: Union[float, int], states: State) -> Update:
        assert states['other'] == 2.0, 'not getting access to other state'
        top = states['top']
        agents = top['agents']

        # update the bigraph directly by dividing agents
        update: Update = {'top': {'agents': {}}}
        for agent_id, agent_state in agents.items():
            internal_glc = agent_state['internal']['GLC']
            if internal_glc >= self.parameters['division_threshold']:
                # trigger division
                daughter_ids = daughter_phylogeny_id(agent_id)
                daughter_updates = [
                    {'key': daughter_id}
                    for daughter_id in daughter_ids]
                update['top']['agents'] = {
                    '_divide': {
                        'mother': agent_id,
                        'daughters': daughter_updates}}

        # examine agent processes
        for agent_id, agent_state in agents.items():
            transport = agent_state['transport']
            transport_process = transport[0]
            transport_topology = transport[1]
            assert isinstance(transport_process, Process)
            assert isinstance(transport_topology, dict)

        return update


def test_bigraph_view() -> None:
    agent_id = '1'

    top_view_steps = {'top_view': TopView()}
    top_view_flow: Flow = {'top_view': []}
    top_view_topology: Topology = {
        'top_view': {
            'top': (),  # connect to the top
            'other': ('other',),
        }
    }

    composite = get_toy_transport_in_env_composite(agent_id=agent_id)
    composite.merge(
        steps=top_view_steps,
        topology=top_view_topology,
        flow=top_view_flow)

    # run the simulation
    sim = Engine(
        composite=composite,
        initial_state={
            'agents': {agent_id: {'external': {'GLC': 10.0}}}}
    )
    sim.update(20)
    data = sim.emitter.get_data()

    print(pf(data))
    len(data[20.0]['agents'])


scans_library = {
    '0': test_bigraph_view,
}

# python vivarium/experiments/bigraph_view.py -n [name]
if __name__ == '__main__':
    run_library_cli(scans_library)



"""
Experiment to profile runtime in process next_update vs
remaining vivarium overhead

Execute by running:
``python vivarium/experiments/profile_runtime.py -n [scan id]``
"""

import os
import random
import time
import cProfile
import pstats

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from vivarium.core.engine import Engine
from vivarium.core.process import Process
from vivarium.core.composer import Composer
from vivarium.core.control import run_library_cli
from vivarium.core.directories import EXPERIMENT_OUT_DIR

DEFAULT_PROCESS_SLEEP = 1e-3
DEFAULT_N_PROCESSES = 10
DEFAULT_N_VARIABLES = 10
DEFAULT_EXPERIMENT_TIME = 100
PROCESS_UPDATE_MARKER = 'bD'
VIVARIUM_OVERHEAD_MARKER = 'rD'
SIMULATION_TIME_MARKER = 'gD'


class ManyVariablesProcess(Process):
    defaults = {
        'number_of_ports': 1,
        'number_of_variables': DEFAULT_N_VARIABLES,
        'process_sleep': DEFAULT_PROCESS_SLEEP,
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)
        # make a bunch of port and variable ids
        port_ids = [*range(
            self.parameters['number_of_ports'])]
        variable_ids = [*range(
            self.parameters['number_of_variables'])]

        # assign variables to each port
        self.port_variables = {port_id: [] for port_id in port_ids}
        port_idx = 0
        for var_id in variable_ids:
            port_id = port_ids[port_idx]
            self.port_variables[port_id].append(var_id)
            if port_idx >= (self.parameters['number_of_ports'] - 1):
                port_idx = 0
            else:
                port_idx += 1

    def ports_schema(self):
        ports = {
            port_id: {
                variable: {
                    '_default': random.random(),
                    '_emit': True
                } for variable in variable_ids
            } for port_id, variable_ids in self.port_variables.items()}
        return ports

    def next_update(self, timestep, states):
        update = {}
        for port_id, variable_ids in self.port_variables.items():
            port_update = {
                variable: random.random()
                for variable in variable_ids}
            update[port_id] = port_update

        # use while loop for busy wait, to use CPU time
        current_time = time.time()
        while time.time() < current_time + self.parameters['process_sleep']:
            pass

        return update


class ManyVariablesComposite(Composer):
    defaults = {
        'number_of_processes': DEFAULT_N_PROCESSES,
        'number_of_variables': DEFAULT_N_VARIABLES,
        'process_sleep': DEFAULT_PROCESS_SLEEP,
        'number_of_parallel_processes': 0,
        'number_of_stores': 10,
        'number_of_ports': 1,
        'hierarchy_depth': 1,
    }

    def __init__(self, config=None):
        super().__init__(config)
        process_ids = [
            f'process_{key}' for key
            in range(self.config['number_of_processes'])]
        parallel_process_ids = [
            process_id for process_idx, process_id in enumerate(process_ids, 1)
            if process_idx <= self.config['number_of_parallel_processes']]
        store_ids = [
            f'store_{key}' for key
            in range(self.config['number_of_stores'])]

        # make a bunch of processes
        self.processes = {
            process_id: ManyVariablesProcess({
                'name': process_id,
                'number_of_ports': self.config['number_of_ports'],
                'number_of_variables': self.config['number_of_variables'],
                'process_sleep': self.config['process_sleep'],
                '_parallel': process_id in parallel_process_ids,
            })
            for process_id in process_ids}

        # connect the processes' ports to a random store at hierarchy_depth
        # TODO -- control number of stores at hierarchy depth
        self.topology = {}
        for process_id, process in self.processes.items():
            process_ports = {}
            for port_id in process.port_variables.keys():
                store_id = random.choice(store_ids)
                store_path = [
                    store_id for _
                    in range(self.config['hierarchy_depth'])]
                process_ports[port_id] = tuple(store_path)
            self.topology[process_id] = process_ports

    def generate_processes(self, config):
        return self.processes

    def generate_topology(self, config):
        return self.topology


class ComplexModelSim:
    """Profile Complex Models

    This class lets you initialize and profile the simulation of
    composite models with arbitrary numbers of processes, variables
    per process, and total stores.
    """

    # model complexity
    number_of_processes = DEFAULT_N_PROCESSES
    number_of_variables = DEFAULT_N_VARIABLES
    process_sleep = DEFAULT_PROCESS_SLEEP
    number_of_parallel_processes = 0
    number_of_stores = 10
    number_of_ports = 1
    hierarchy_depth = 1
    experiment_time = DEFAULT_EXPERIMENT_TIME

    # display
    print_top_stats = 4

    # initialize
    composite = None
    experiment = None

    def set_parameters(
            self,
            number_of_processes=None,
            number_of_parallel_processes=None,
            number_of_stores=None,
            number_of_ports=None,
            number_of_variables=None,
            hierarchy_depth=None,
            process_sleep=None,
            print_top_stats=None,
            experiment_time=None,
    ):
        self.number_of_processes = \
            number_of_processes or self.number_of_processes
        self.number_of_parallel_processes = \
            number_of_parallel_processes or self.number_of_parallel_processes
        self.number_of_ports = \
            number_of_ports or self.number_of_ports
        self.number_of_variables = \
            number_of_variables or self.number_of_variables
        self.number_of_stores = \
            number_of_stores or self.number_of_stores
        self.hierarchy_depth = \
            hierarchy_depth or self.hierarchy_depth
        self.process_sleep = \
            process_sleep or self.process_sleep
        self.print_top_stats = \
            print_top_stats or self.print_top_stats
        self.experiment_time = \
            experiment_time or self.experiment_time

    def _generate_composite(self, **kwargs):
        number_of_processes = kwargs.get(
            'number_of_processes', self.number_of_processes)
        number_of_parallel_processes = kwargs.get(
            'number_of_parallel_processes', self.number_of_parallel_processes)
        number_of_stores = kwargs.get(
            'number_of_stores', self.number_of_stores)
        number_of_ports = kwargs.get(
            'number_of_ports', self.number_of_ports)
        number_of_variables = kwargs.get(
            'number_of_variables', self.number_of_variables)
        hierarchy_depth = kwargs.get(
            'hierarchy_depth', self.hierarchy_depth)
        process_sleep = kwargs.get(
            'process_sleep', self.process_sleep)

        composer = ManyVariablesComposite({
            'number_of_processes': number_of_processes,
            'number_of_parallel_processes': number_of_parallel_processes,
            'number_of_stores': number_of_stores,
            'number_of_ports': number_of_ports,
            'number_of_variables': number_of_variables,
            'hierarchy_depth': hierarchy_depth,
            'process_sleep': process_sleep,
        })

        self.composite = composer.generate(**kwargs)

    def _initialize_experiment(self, **kwargs):
        self.experiment = Engine(
            processes=self.composite['processes'],
            topology=self.composite['topology'],
            **kwargs)

    def _run_experiment(self, **kwargs):
        self.experiment.update(kwargs['experiment_time'])
        self.experiment.end()

    def _get_emitter_data(self, **kwargs):
        _ = kwargs
        data = self.experiment.emitter.get_data()
        return data

    def _get_emitter_timeseries(self, **kwargs):
        _ = kwargs
        timeseries = self.experiment.emitter.get_timeseries()
        return timeseries

    def _profile_method(self, method, **kwargs):
        """The main profiling method and of the simulation steps

        Args
            method: the simulation step. For example self._run_experiment
        """
        print_top_stats = kwargs.get(
            'print_top_stats', self.print_top_stats)
        profiler = cProfile.Profile()
        profiler.enable()
        method(**kwargs)
        profiler.disable()
        stats = pstats.Stats(profiler)
        if print_top_stats:
            stats.sort_stats('tottime').print_stats(print_top_stats)
        return stats

    def profile_communication_latency(self):

        self._generate_composite()
        self._initialize_experiment(display_info=False)

        # profile the experiment
        stats = self._profile_method(
            self._run_experiment,
            experiment_time=self.experiment_time,
            print_top_stats=None)

        # get next_update runtime
        next_update_amount = ("next_update",)
        _, stats_list = stats.get_print_list(next_update_amount)

        process_update_time = 0
        for s in stats_list:
            process_update_time += stats.stats[s][3]

        # get runtime
        experiment_time = stats.total_tt
        store_update_time = experiment_time - process_update_time

        return process_update_time, store_update_time


# Parameter scan functions
##########################

def run_scan(
    sim,
    scan_values=None,
):
    """Run a scan

    Args
        sim: the ComplexModelSim object.
        scan_values: a list of dicts, with individual scan values.
    """
    scan_values = scan_values or []

    saved_stats = []
    for scan_dict in scan_values:
        # set the parameters
        sim.set_parameters(
            number_of_processes=scan_dict.get('number_of_processes'),
            process_sleep=scan_dict.get('process_sleep'),
            number_of_parallel_processes=scan_dict.get(
                'number_of_parallel_processes'),
            number_of_stores=scan_dict.get('number_of_stores'),
            number_of_ports=scan_dict.get('number_of_ports'),
            number_of_variables=scan_dict.get('number_of_variables'),
            hierarchy_depth=scan_dict.get('hierarchy_depth'),
        )

        print(f"{scan_dict}")

        # run experiment
        process_update_time, store_update_time = \
            sim.profile_communication_latency()

        stat_dict = {
            **scan_dict,
            'process_update_time': process_update_time,
            'store_update_time': store_update_time,
        }
        saved_stats.append(stat_dict)

    return saved_stats


# Plotting functions
####################

def _make_axis(fig, grid, plot_n, patches, title='', label=''):
    ax = fig.add_subplot(grid[plot_n, 0])
    ax.set_xlabel(label)
    ax.set_ylabel('wall time (s)')
    ax.set_title(title)
    ax.legend(
        handles=patches,
        # ncol=2,
        # bbox_to_anchor=(0.5, 1.2),  # above
        bbox_to_anchor=(1.45, 0.65),  # to the right
    )
    return ax


def _get_patches(
        process=False,
        overhead=False,
        experiment=False
):
    patches = []
    if process:
        patches.append(mpatches.Patch(
            color=PROCESS_UPDATE_MARKER[0],
            label="process updates"))
    if overhead:
        patches.append(mpatches.Patch(
            color=VIVARIUM_OVERHEAD_MARKER[0],
            label="vivarium overhead"))
    if experiment:
        patches.append(mpatches.Patch(
            color=SIMULATION_TIME_MARKER[0],
            label="simulation time"))
    return patches


def _add_stats_plot(
        ax,
        saved_stats,
        variable_name,
        process_update=False,
        vivarium_overhead=False,
        experiment_time=False,
        markersize=10,
):
    # plot saved states
    for stat in saved_stats:
        variable = stat[variable_name]
        process_update_time = stat['process_update_time']
        store_update_time = stat['store_update_time']

        if process_update:
            ax.plot(
                variable, process_update_time,
                PROCESS_UPDATE_MARKER, markersize=markersize)
        if vivarium_overhead:
            ax.plot(
                variable, store_update_time,
                VIVARIUM_OVERHEAD_MARKER, markersize=markersize)
        if experiment_time:
            experiment_time = process_update_time + store_update_time
            ax.plot(
                variable, experiment_time,
                SIMULATION_TIME_MARKER, markersize=markersize)


def plot_scan_results(
        saved_stats,
        process_plot=False,
        store_plot=False,
        port_plot=False,
        var_plot=False,
        hierarchy_plot=False,
        parallel_plot=False,
        cpus_plot=False,
        fig=None,
        grid=None,
        axis_number=0,
        row_height=3,
        title=None,
        out_dir=EXPERIMENT_OUT_DIR,
        filename=None,
):
    """Plot scan results

    Args
        saved_stats (dict): the scan results
        *_plot (bool): whether to add the given plot type
    """
    # make figure
    if fig:
        assert grid, "fig must provide grid for subplots"
    else:
        n_cols = 1
        n_rows = sum([
            process_plot,
            store_plot,
            port_plot, port_plot,  # two rows
            var_plot, var_plot,  # two rows
            hierarchy_plot,
            parallel_plot,
            cpus_plot,
        ])

        fig = plt.figure(figsize=(n_cols * 6, n_rows * row_height))
        grid = plt.GridSpec(n_rows, n_cols)

    patches = _get_patches(
        process=True, overhead=True, experiment=True)

    # initialize axes
    if process_plot:
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label='n processes')
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name='number_of_processes',
            process_update=True,
            vivarium_overhead=True)
        axis_number += 1

    if store_plot:
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label='n stores')
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name='number_of_stores',
            vivarium_overhead=True)
        axis_number += 1

    if port_plot:
        label = 'n ports'
        var_name = 'number_of_ports'
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label=label)
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name=var_name,
            # process_update=True,
            vivarium_overhead=True)
        axis_number += 1

        # process time plot
        ax = _make_axis(
            fig, grid, axis_number, patches, title='',
            label=label)
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name=var_name,
            process_update=True)
        axis_number += 1

    if var_plot:
        label = 'n variables'
        var_name = 'number_of_variables'
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label=label)
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name=var_name,
            vivarium_overhead=True)
        axis_number += 1

        # process time plot
        ax = _make_axis(
            fig, grid, axis_number, patches, title='',
            label=label)
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name=var_name,
            process_update=True)
        axis_number += 1

    if hierarchy_plot:
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label='hierarchy depth')
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name='hierarchy_depth',
            vivarium_overhead=True)
        axis_number += 1

    if parallel_plot:
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label='n parallel processes')
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name='number_of_parallel_processes',
            experiment_time=True)
        axis_number += 1

    if cpus_plot:
        ax = _make_axis(
            fig, grid, axis_number, patches, title,
            label='n vCPUs')
        _add_stats_plot(
            ax=ax, saved_stats=saved_stats,
            variable_name='number_of_cpus',
            experiment_time=True)
        axis_number += 1

    # save
    if filename:
        plt.subplots_adjust(hspace=0.5)
        # plt.figtext(0, -0.1, filename, size=8)
        os.makedirs(out_dir, exist_ok=True)
        fig_path = os.path.join(out_dir, filename[0:100])
        fig.savefig(fig_path, bbox_inches='tight')
    return fig


# Individual scan functions
###########################

def scan_processes():
    n_processes = [n*15 for n in range(10)]
    sleep_times = [0.8e-5, 0.8e-4, 0.8e-3]

    n_cols = 1
    n_rows = len(sleep_times)
    fig = plt.figure(figsize=(n_cols * 6, n_rows * 3))
    grid = plt.GridSpec(n_rows, n_cols)

    for idx, s in enumerate(sleep_times):
        scan_values = []
        for n in n_processes:
            scan_value = {
                'number_of_processes': n,
                'process_sleep': s,
            }
            scan_values.append(scan_value)

        sim = ComplexModelSim()
        saved_stats = run_scan(sim,
                               scan_values=scan_values)

        plot_scan_results(saved_stats,
                          process_plot=True,
                          row_height=2.5,
                          fig=fig,
                          grid=grid,
                          axis_number=idx,
                          title=f'process update runtime = {s} sec',
                          # filename='scan_processes'
                          )
    plot_scan_results({},
                      fig=fig,
                      grid=grid,
                      filename='scan_processes'
                      )


def scan_variables():
    n_vars = [n*100 for n in range(10)]
    scan_values = [{'number_of_variables': n} for n in n_vars]

    sim = ComplexModelSim()
    saved_stats = run_scan(sim,
                           scan_values=scan_values)
    plot_scan_results(saved_stats,
                      var_plot=True,
                      row_height=1.8,
                      filename='scan_variables',
                      title='n variables through 1 port'
                      )


def scan_number_of_ports():
    n_ports = [n*10 for n in range(10)]
    scan_values = [
        {
            'number_of_variables': 100,
            'number_of_ports': n
        } for n in n_ports
    ]

    sim = ComplexModelSim()
    saved_stats = run_scan(sim,
                           scan_values=scan_values)
    plot_scan_results(saved_stats,
                      port_plot=True,
                      row_height=1.8,
                      filename='scan_number_of_ports',
                      title='100 variables through n ports'
                      )


def scan_hierarchy_depth():
    hierarchy_depth = [n*2 for n in range(20)]
    scan_values = [{'hierarchy_depth': n} for n in hierarchy_depth]

    sim = ComplexModelSim()
    saved_stats = run_scan(sim,
                           scan_values=scan_values)
    plot_scan_results(saved_stats,
                      hierarchy_plot=True,
                      filename='scan_hierarchy_depth')


def scan_parallel_processes():
    """
    Running this scan may require change the computer's process limits.
     - Check your current limits with: `ulimit -n`
     - Increase limits: `ulimit -n 10240`
    """

    total_processes = 80
    scan_interval = 10
    scan_values = [
        {
            'number_of_processes': total_processes,
            'number_of_parallel_processes': n
        } for n in range(0, total_processes+1, scan_interval)
    ]

    sim = ComplexModelSim()
    sim.process_sleep = 1e-2
    saved_stats = run_scan(sim,
                           scan_values=scan_values)
    plot_scan_results(saved_stats,
                      parallel_plot=True,
                      row_height=2.5,
                      title=f'{total_processes} processes, '
                            f'with n of them running in parallel',
                      filename='scan_parallel_processes')


scans_library = {
    '0': scan_processes,
    '1': scan_variables,
    '2': scan_number_of_ports,
    '3': scan_parallel_processes,
    '4': scan_hierarchy_depth,
}

# python vivarium/experiments/profile_runtime.py -n [name]
if __name__ == '__main__':
    run_library_cli(scans_library)



"""
================
Process Classes
================
"""

import abc
import copy
import cProfile
from multiprocessing import Pipe
from multiprocessing import Process as Multiprocess
from multiprocessing.connection import Connection
import os
import pstats
import pickle
from typing import Any, Dict, Optional, Union, List, Tuple
from warnings import warn

import pytest

from vivarium.library.dict_utils import (
    deep_merge, deep_merge_check, deep_copy_internal)
from vivarium.library.topology import assoc_path, get_in
from vivarium.core.types import (
    HierarchyPath, Schema, State, Update,
    Topology, Flow)

DEFAULT_TIME_STEP = 1.0


def assoc_in(d: dict, path: HierarchyPath, value: Any) -> dict:
    """Insert a value into a dictionary at an arbitrary depth.

    Empty dictionaries will be created as needed to insert the value at
    the specified depth.

    >>> d = {'a': {'b': 1}}
    >>> assoc_in(d, ('a', 'c', 'd'), 2)
    {'a': {'b': 1, 'c': {'d': 2}}}

    Args:
        d: Dictionary to insert into.
        path: Path in the dictionary where the value will be inserted.
            Each element of the path is dictionary key, which will be
            added if not already present. Any given element (except the
            first) refers to a key in the dictionary that is the value
            associated with the immediately preceding path element.
        value: The value to insert.

    Returns:
        Dictionary with the value inserted.
    """
    if path:
        return dict(
            d,
            **{
                path[0]: assoc_in(d.get(path[0], {}), path[1:], value)
            }
        )
    return value


def _override_schemas(
        overrides: Dict[str, Schema],
        processes: Dict[str, 'Process']
) -> None:
    for key, override in overrides.items():
        process = processes[key]
        if isinstance(process, Process):
            process.merge_overrides(override)
        elif isinstance(process, dict):
            _override_schemas(override, process)


def _get_parameters(
        processes: Optional[Dict[str, 'Process']] = None
) -> Dict:
    processes = processes or {}
    parameters: Dict = {}
    for key, value in processes.items():
        if isinstance(value, Process):
            parameters[key] = value.parameters
        elif isinstance(value, dict):
            parameters[key] = _get_parameters(value)
    return parameters


class Process(metaclass=abc.ABCMeta):
    """Process parent class.

      All :term:`process` classes must inherit from this class. Each
      class can provide a ``defaults`` class variable to specify the
      process defaults as a dictionary.

      Note that subclasses should call the superclass init function
      first. This allows the superclass to correctly save the initial
      parameters before they are mutated by subclass constructor code.
      We need access to the original parameters for serialization to
      work properly.

      Args:
          parameters: Override the class defaults. This dictionary may
              also contain the following special keys:

              * ``name``: Saved to ``self.name``.
              * ``_schema``: Overrides the schema.
              * ``_parallel``: Indicates that the process should be
                parallelized. ``self.parallel`` will be set to True.
              * ``_condition``: Path to a variable whose value will be
                returned by
                :py:meth:`vivarium.core.process.Process.update_condition`.
              * ``time_step``: Returned by
                :py:meth:`vivarium.core.process.Process.calculate_timestep`.
      """

    defaults: Dict[str, Any] = {}
    METHOD_COMMANDS = (
        'initial_state', 'generate_processes', 'generate_steps',
        'generate_topology', 'generate_flow', 'merge_overrides',
        'calculate_timestep', 'is_step', 'get_private_state',
        'ports_schema', 'next_update', 'update_condition')
    ATTRIBUTE_READ_COMMANDS = (
        'schema_override', 'parameters', 'condition_path', 'schema')
    ATTRIBUTE_WRITE_COMMANDS = ('set_schema',)

    def __init__(self, parameters: Optional[dict] = None) -> None:
        parameters = parameters or {}

        if 'name' in parameters:
            self.name = parameters['name']
        elif not hasattr(self, 'name'):
            self.name = self.__class__.__name__

        self._parameters = copy.deepcopy(self.defaults)
        self._parameters = deep_merge(self._parameters, parameters)
        self._schema_override: Schema = self._parameters.get('_schema', {})
        self._parallel = self._parameters.get('_parallel', False)
        self._condition_path: Optional[HierarchyPath] = None
        self._command_result: Any = None
        self._pending_command: Optional[
            Tuple[str, Optional[tuple], Optional[dict]]] = None

        # set up the conditional state if a condition key is provided
        if '_condition' in self._parameters:
            self._condition_path = self._parameters.get('_condition')
        if self._condition_path:
            self.merge_overrides(assoc_path({}, self._condition_path, {
                '_default': True,
                '_emit': True,
                '_updater': 'set'}))

        self._set_timestep()

        self._schema: Optional[Schema] = None

    @property
    def parameters(self) -> dict:
        return self._parameters

    @property
    def schema_override(self) -> Schema:
        return self._schema_override

    @property
    def parallel(self) -> bool:
        return self._parallel

    @property
    def condition_path(self) -> Optional[HierarchyPath]:
        return self._condition_path

    @property
    def schema(self) -> Optional[Schema]:
        return self._schema

    @schema.setter
    def schema(self, value: Optional[Schema]) -> None:
        self._schema = value

    def pre_send_command(
            self, command: str, args: Optional[tuple], kwargs:
            Optional[dict]) -> None:
        '''Run pre-checks before starting a command.

        This method should be called at the start of every
        implementation of :py:meth:`send_command`.

        Args:
            command: The name of the command to run.
            args: A tuple of positional arguments for the command.
            kwargs: A dictionary of keyword arguments for the command.

        Raises:
            RuntimeError: Raised when a user tries to send a command
                while a previous command is still pending (i.e. the user
                hasn't called :py:meth:`get_command_result` yet for the
                previous command).
        '''
        if self._pending_command:
            raise RuntimeError(
                f'Trying to send command {(command, args, kwargs)} but '
                f'command {self._pending_command} is still pending.')
        self._pending_command = command, args, kwargs


    def send_command(
            self, command: str, args: Optional[tuple] = None,
            kwargs: Optional[dict] = None,
            run_pre_check: bool = True) -> None:
        '''Handle :term:`process commands`.

        This method handles the commands listed in
        :py:attr:`METHOD_COMMANDS` by passing ``args``
        and ``kwargs`` to the method of ``self`` with the name
        of the command and saving the return value as the result.

        This method handles the commands listed in
        :py:attr:`ATTRIBUTE_READ_COMMANDS` by returning the attribute of
        ``self`` with the name matching the command, and it handles the
        commands listed in :py:attr:`ATTRIBUTE_WRITE_COMMANDS` by
        setting the attribute in the command to the first argument in
        ``args``. The command must be named ``set_attr`` for attribute
        ``attr``.

        To add support for a custom command, override this function in
        your subclass. Each command is defined by a name (a string)
        and accepts both positional and keyword arguments. Any custom
        commands you add should have associated methods such that:

        * The command name matches the method name.
        * The command and method accept the same positional and keyword
          arguments.
        * The command and method return the same values.

        If all of the above are satisfied, you can use
        :py:meth:`Process.run_command_method` to handle the command.

        Your implementation of this function needs to handle all the
        commands you want to support.  When presented with an unknown
        command, you should call the superclass method, which will
        either handle the command or call its superclass method. At the
        top of this recursive chain, this ``Process.send_command()``
        method handles some built-in commands and will raise an error
        for unknown commands.

        Any overrides of this method must also call
        :py:meth:`pre_send_command` at the start of the method. This
        call will check that no command is currently pending to avoid
        confusing behavior when multiple commands are started without
        intervening retrievals of command results. Since your overriding
        method will have already performed the pre-check, it should pass
        ``run_pre_check=False`` when calling the superclass method.

        Args:
            command: The name of the command to run.
            args: A tuple of positional arguments for the command.
            kwargs: A dictionary of keyword arguments for the command.
            run_pre_check: Whether to run the pre-checks implemented in
                :py:meth:`pre_send_command`. This should be left at its
                default value unless the pre-checks have already been
                performed (e.g. if this method is being called by a
                subclass's overriding method.)

        Returns:
            None. This method just starts the command running.

        Raises:
            ValueError: For unknown commands.
        '''
        if run_pre_check:
            self.pre_send_command(command, args, kwargs)
        args = args or tuple()
        kwargs = kwargs or {}
        if command in self.METHOD_COMMANDS:
            self._command_result = self.run_command_method(
                command, args, kwargs)
        elif command in self.ATTRIBUTE_READ_COMMANDS:
            self._command_result = getattr(self, command)
        elif command in self.ATTRIBUTE_WRITE_COMMANDS:
            assert command.startswith('set_')
            assert args
            setattr(self, command[len('set_'):], args[0])
        else:
            raise ValueError(
                f'Process {self} does not understand the process '
                f'command {command}')

    def run_command_method(
            self, command: str, args: tuple, kwargs: dict) -> Any:
        '''Run a command whose name and interface match a method.

        Args:
            command: The command name, which must equal to a method of
                ``self``.
            args: The positional arguments to pass to the method.
            kwargs: The keywords arguments for the method.

        Returns:
            The result of calling ``self.command(*args, **kwargs)`` is
            returned for command ``command``.
        '''
        return getattr(self, command)(*args, **kwargs)

    def get_command_result(self) -> Any:
        '''Retrieve the result from the last-run command.

        Returns:
            The result of the last command run. Note that this method
            should only be called once immediately after each call to
            :py:meth:`send_command`.

        Raises:
            RuntimeError: When there is no command pending. This can
                happen when this method is called twice without an
                intervening call to :py:meth:`send_command`.
        '''
        if not self._pending_command:
            raise RuntimeError(
                'Trying to retrieve command result, but no command is '
                'pending.')
        self._pending_command = None
        result = self._command_result
        self._command_result = None
        return result

    def run_command(
            self, command: str, args: Optional[tuple] = None,
            kwargs: Optional[dict] = None) -> Any:
        '''Helper function that sends a command and returns result.'''
        self.send_command(command, args, kwargs)
        return self.get_command_result()

    def _set_timestep(self) -> None:
        self._parameters.setdefault('timestep', DEFAULT_TIME_STEP)
        if self._parameters.get('time_step'):
            self._parameters['timestep'] = self._parameters['time_step']


    def initial_state(self, config: Optional[dict] = None) -> State:
        """Get initial state in embedded path dictionary.

        Every subclass may override this method.

        Args:
            config (dict): A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            dict: Subclass implementations must return a dictionary
            mapping state paths to initial values.
        """
        _ = config
        return {}

    def generate_processes(
            self, config: Optional[dict] = None) -> Dict[str, Any]:
        """Do not override this method."""
        if not self.is_step():
            config = config or {}
            name = config.get('name', self.name)
            return {name: self}
        return {}

    def generate_steps(
            self, config: Optional[dict] = None) -> Dict[str, Any]:
        """Do not override this method."""
        if self.is_step():
            config = config or {}
            name = config.get('name', self.name)
            return {name: self}
        return {}

    def generate_topology(self, config: Optional[dict] = None) -> Topology:
        """Do not override this method."""
        config = config or {}
        name = config.get('name', self.name)
        override_topology = config.get('topology', {})
        ports = self.ports()
        return {
            name: {
                port: override_topology.get(port, (port,))
                for port in ports.keys()}}

    def generate_flow(self, config: Optional[dict] = None) -> Flow:
        _ = config
        return {}

    def generate(
            self,
            config: Optional[dict] = None,
            path: HierarchyPath = ()) -> Dict:
        if config is None:
            config = self.parameters
        else:
            default = copy.deepcopy(self.parameters)
            config = deep_merge(default, config)
        config = config or {}
        name = config.get('name', self.name)
        processes = self.generate_processes(config)
        steps = self.generate_steps(config)
        topology = self.generate_topology(config)
        flow = self.generate_flow(config)
        processes_and_steps = deep_copy_internal(processes)
        deep_merge_check(processes_and_steps, steps)
        _override_schemas(
            {name: self.schema_override},
            processes_and_steps)

        # TODO -- this should return a Composite instance,
        # but importing Composite introduces circular imports
        # from vivarium.core.composer import Composite
        return {
            'processes': assoc_in({}, path, processes),
            'steps': assoc_in({}, path, steps),
            'topology': assoc_in({}, path, topology),
            'flow': assoc_in({}, path, flow),
            'state': {}
        }

    def get_schema(self, override: Optional[Schema] = None) -> dict:
        """Get the process's schema, optionally with a schema override.

        Args:
            override: Override schema

        Returns:
            The combined schema.
        """
        ports = copy.deepcopy(self.ports_schema())
        deep_merge(ports, self.schema_override)
        deep_merge(ports, override)
        return ports

    def merge_overrides(self, override: Schema) -> None:
        """Add a schema override to the process's schema overrides.

        Args:
            override: The schema override to add.
        """
        deep_merge(self._schema_override, override)

    def ports(self) -> Dict[str, List[str]]:
        """Get ports and each port's variables.

        Returns:
            A map from port names to lists of the variables that go into
            that port.
        """
        ports_schema = self.ports_schema()
        return {
            port: list(states.keys())
            for port, states in ports_schema.items()}

    def calculate_timestep(self, states: Optional[State]) -> Union[float, int]:
        """Return the next process time step

        A process subclass may override this method to implement
        adaptive timesteps. By default it returns self.parameters['timestep'].
        """
        _ = states
        return self.parameters['timestep']

    def default_state(self) -> State:
        """Get the default values of the variables in each port.

        The default values are computed based on the schema.

        Returns:
            A state dictionary that assigns each variable's default
            value to that variable.
        """
        schema = self.get_schema()
        state: State = {}
        for port, states in schema.items():
            for key, value in states.items():
                if '_default' in value:
                    if port not in state:
                        state[port] = {}
                    state[port][key] = value['_default']
        return state

    def is_deriver(self) -> bool:
        """Check whether this process is a :term:`deriver`.

        .. deprecated:: 0.3.14
           Derivers have been deprecated in favor of :term:`steps`, so
           please override ``is_step`` instead of ``is_deriver``.
           Support for Derivers may be removed in a future release.

        Returns:
            Whether this process is a deriver. This class always returns
            ``False``, but subclasses may change this.
        """
        return False

    def is_step(self) -> bool:
        """Check whether this process is a :term:`step`.

        Returns:
            Whether this process is a step. This class always returns
            ``False``, but subclasses may change this behavior.
        """
        method = getattr(self.is_deriver, '__func__', None)
        if method and method is not Process.is_deriver:
            # `self` is an instance of a subclass that has overridden
            # `is_deriver`.
            warn(
                'is_deriver() is deprecated. Use is_step() instead.',
                category=FutureWarning)
            return self.is_deriver()
        return False

    def get_private_state(self) -> State:
        """Get the process's private state.

        Processes can store state in instance variables instead of in
        the :term:`stores` that hold the simulation-wide state. These
        instance variables hold a private state that is not shared with
        other processes. You can override this function to let
        :term:`experiments` emit the process's private state.

        Returns:
            An empty dictionary. You may override this behavior to
            return your process's private state.
        """
        return {}  # pragma: no cover

    @abc.abstractmethod
    def ports_schema(self) -> Schema:
        """Get the schemas for each port.

        This must be overridden by any subclasses.

        Use the glob '*' schema to declare expected sub-store structure,
        and view all child values of the store:

          .. code-block:: python

            schema = {
                'port1': {
                    '*': {
                        '_default': 1.0
                    }
                }
            }

        Use the glob '**' schema to connect to an entire sub-branch, including
        child nodes, grandchild nodes, etc:

          .. code-block:: python

            schema = {
                'port1': '**'
            }

        Ports flagged as output-only won't be viewed through the next_update's
        states, which can save some overhead time:

          .. code-block:: python

            schema = {
              'port1': {
                 '_output': True,
                 'A': {'_default': 1.0},
              }
            }

        Returns:
            A dictionary that declares which states are expected by the
            processes, and how each state will behave. State keys can be
            assigned properties through schema_keys declared in
            :py:class:`vivarium.core.store.Store`.
        """
        return {}  # pragma: no cover

    @abc.abstractmethod
    def next_update(
            self, timestep: Union[float, int], states: State) -> Update:
        """Compute the next update to the simulation state.

        Args:
            timestep: The duration for which the update should be
                computed.
            states: The pre-update simulation state. This will take the
                same form as the process's schema, except with a value
                for each variable.

        Returns:
            An empty dictionary for now. This should be overridden by
            each subclass to return an update.
        """
        return {}  # pragma: no cover

    def update_condition(
            self, timestep: Union[float, int], states: State) -> Any:
        """Determine whether this process runs.

        Args:
            timestep: The duration for which an update.
            states: The pre-update simulation state.

        Returns:
            Boolean for whether this process runs. True by default.
        """

        _ = timestep
        _ = states

        # use the given condition key if it was provided
        if self.condition_path:
            return get_in(states, self.condition_path)

        return True


class Step(Process, metaclass=abc.ABCMeta):
    """Base class for steps."""

    def __init__(self, parameters: Optional[dict] = None) -> None:
        parameters = parameters or {}
        super().__init__(parameters)

    def is_step(self) -> bool:
        """Returns ``True`` to signal that this process is a step."""
        return True


#: Deriver is just an alias for :py:class:`Step` now that Derivers have
#: been deprecated.
Deriver = Step


def _handle_parallel_process(
        connection: Connection, process: Process,
        profile: bool) -> None:
    '''Handle a parallel Vivarium :term:`process`.

    This function is designed to be passed as ``target`` to
    ``Multiprocess()``. In a loop, it receives :term:`process commands`
    from a pipe, passes those commands to the parallel process, and
    passes the result back along the pipe.

    The special command ``end`` is handled directly by this function.
    This command causes the function to exit and therefore shut down the
    OS process created by multiprocessing.

    Args:
        connection: The child end of a multiprocessing pipe. All
            communications received from the pipe should be a 3-tuple of
            the form ``(command, args, kwargs)``, and the tuple contents
            will be passed to :py:meth:`Process.run_command`. The
            result, which may be of any type, will be sent back through
            the pipe.
        process: The process running in parallel.
        profile: Whether to profile the process.
    '''
    if profile:
        profiler = cProfile.Profile()
        profiler.enable()
    running = True

    while running:
        command, args, kwargs = connection.recv()

        if command == 'end':
            running = False
        else:
            result = process.run_command(command, args, kwargs)
            connection.send(result)

    if profile:
        profiler.disable()
        stats = pstats.Stats(profiler)
        connection.send(stats.stats)  # type: ignore

    connection.close()


class ParallelProcess(Process):
    def __init__(
            self, process: Process, profile: bool = False,
            stats_objs: Optional[List[pstats.Stats]] = None) -> None:
        """Wraps a :py:class:`Process` for multiprocessing.

        To run a simulation distributed across multiple processors, we
        use Python's multiprocessing tools. This object runs in the main
        process and manages communication between the main (parent)
        process and the child process with the :py:class:`Process` that
        this object manages.

        Most methods pass their name and arguments to
        :py:class:`Process.run_command`.

        Args:
            process: The Process to manage.
            profile: Whether to use cProfile to profile the subprocess.
            stats_objs: List to add cProfile stats objs to when process
                is deleted. Only used if ``profile`` is true.
        """
        super().__init__({
            'name': process.name,
            '_parallel': True,
        })
        self.profile = profile
        self._stats_objs = stats_objs
        assert not self.profile or self._stats_objs is not None
        self.parent, child = Pipe()
        self.multiprocess = Multiprocess(
            target=_handle_parallel_process,
            args=(child, process, self.profile))
        self.multiprocess.start()
        self._ended = False
        self._pending_command: Optional[
            Tuple[str, Optional[tuple], Optional[dict]]] = None

    def send_command(
            self, command: str, args: Optional[tuple] = None,
            kwargs: Optional[dict] = None,
            run_pre_check: bool = True) -> None:
        '''Send a command to the parallel process.

        See :py:func:``_handle_parallel_process`` for details on how the
        command will be handled.
        '''
        if run_pre_check:
            self.pre_send_command(command, args, kwargs)
        self.parent.send((command, args, kwargs))

    def get_command_result(self) -> Update:
        """Get the result of a command sent to the parallel process.

        Commands and their results work like a queue, so unlike
        :py:class:`Process`, you can technically call this method
        multiple times and get different return values each time.
        This behavior is subject to change, so you should not rely on
        it.

        Returns:
            The command result.
        """
        if not self._pending_command:
            raise RuntimeError(
                'Trying to retrieve command result, but no command is '
                'pending.')
        self._pending_command = None
        return self.parent.recv()

    def initial_state(self, config: Optional[dict] = None) -> State:
        return self.run_command('initial_state', (config,))

    def generate_processes(
            self, config: Optional[dict] = None) -> Dict[str, Any]:
        return self.run_command('generate_processes', (config,))

    def generate_steps(
            self, config: Optional[dict] = None) -> Dict[str, Any]:
        return self.run_command('generate_steps', (config,))

    def generate_topology(
            self, config: Optional[dict] = None) -> Topology:
        return self.run_command('generate_topology', (config,))

    def generate_flow(self, config: Optional[dict] = None) -> Flow:
        return self.run_command('generate_flow', (config,))

    @property
    def schema_override(self) -> Schema:
        return self.run_command('schema_override')

    @property
    def parameters(self) -> Dict[str, Any]:
        return self.run_command('parameters')

    @property
    def condition_path(self) -> Optional[HierarchyPath]:
        return self.run_command('condition_path')

    @property
    def schema(self) -> Schema:
        return self.run_command('schema')

    @schema.setter
    def schema(self, value: Schema) -> None:
        self.run_command('set_schema', (value,))

    def merge_overrides(self, override: Schema) -> None:
        self.run_command('merge_overrides', (override,))

    def calculate_timestep(self, states: Optional[State]) -> float:
        return self.run_command('calculate_timestep', (states,))

    def is_step(self) -> bool:
        return self.run_command('is_step')

    def get_private_state(self) -> State:
        return self.run_command('get_private_state')

    def ports_schema(self) -> Schema:
        return self.run_command('ports_schema')

    def next_update(self, timestep: float, states: State) -> Update:
        return self.run_command('next_update', (timestep, states))

    def update_condition(self, timestep: float, states: State) -> bool:
        return self.run_command('update_condition', (timestep, states))

    def end(self) -> None:
        """End the child process.

        If profiling was enabled, then when the child process ends, it
        will compile its profiling stats and send those to the parent.
        The parent then saves those stats in ``self.stats``.
        """
        # Only end once.
        if self._ended:
            return
        self.send_command('end')
        if self.profile:
            stats = pstats.Stats()
            stats.stats = self.get_command_result()  # type: ignore
            assert self._stats_objs is not None
            self._stats_objs.append(stats)
        self.multiprocess.join()
        self.multiprocess.close()
        self._ended = True

    def __del__(self) -> None:
        self.end()


class ToySerializedProcess(Process):

    defaults: Dict[str, list] = {
        'list': [],
    }

    def __init__(self, parameters: Optional[dict] = None) -> None:
        super().__init__(parameters)
        self.parameters['list'].append(1)

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        return {}


class ToySerializedProcessInheritance(Process):

    defaults = {
        '1': 1,
    }

    def __init__(self, parameters: Optional[dict] = None) -> None:
        parameters = parameters or {}
        super().__init__({
            '2': parameters['1'],
        })

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        return {}


class ToyParallelProcess(Process):

    def compare_pid(self, pid: float) -> bool:
        return os.getpid() == pid

    def send_command(
            self, command: str, args: Optional[tuple] = None,
            kwargs: Optional[dict] = None,
            run_pre_check: bool = True) -> None:
        if run_pre_check:
            self.pre_send_command(command, args, kwargs)
        args = args or tuple()
        kwargs = kwargs or {}
        if command == 'compare_pid':
            self._command_result = self.compare_pid(*args, **kwargs)
        else:
            super().send_command(command, args, kwargs, False)

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        return {}


def test_serialize_process() -> None:
    proc = ToySerializedProcess()
    proc_pickle = pickle.loads(pickle.dumps(proc))

    assert proc.parameters['list'] == [1]
    assert proc_pickle.parameters['list'] == [1]


def test_serialize_process_inheritance() -> None:
    a = ToySerializedProcessInheritance({'1': 0})
    a2 = pickle.loads(pickle.dumps(a))
    assert a2.parameters['2'] == 0


def test_process_commands_pending_safeguard() -> None:
    process = ToySerializedProcess()
    process.send_command('calculate_timestep', (None,))
    with pytest.raises(RuntimeError) as exception:
        process.send_command('next_update', (1, {}))
    msg = "command ('calculate_timestep', (None,), None) is still pending"
    assert msg in str(exception.value)


def test_parallel_process_commands_pending_safeguard() -> None:
    process = ParallelProcess(ToySerializedProcess())
    process.send_command('calculate_timestep', (None,))
    with pytest.raises(RuntimeError) as exception:
        process.send_command('next_update', (1, {}))
    msg = "command ('calculate_timestep', (None,), None) is still pending"
    assert msg in str(exception.value)
    # Reset Process._pending_command so that no warning is thrown when
    # __del__() sends the 'end' command.
    process.get_command_result()


def test_parallel_commands() -> None:
    proc = ToyParallelProcess()
    parallel_proc = ParallelProcess(proc)

    assert proc.compare_pid(os.getpid())
    proc.send_command('compare_pid', (os.getpid(),))
    assert proc.get_command_result()

    parallel_proc.send_command('compare_pid', (os.getpid(),))
    assert not parallel_proc.get_command_result()


def test_invalid_command() -> None:
    proc = ToyParallelProcess()
    with pytest.raises(ValueError) as exception:
        proc.send_command('missing_command')
    msg = 'does not understand the process command missing_command'
    assert msg in str(exception.value)



'''
Execute by running: ``python vivarium/process/template_process.py``

TODO: Replace the template code to implement your own process.
'''

import os

from vivarium.core.process import Process
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.core.engine import Engine
from vivarium.plots.simulation_output import plot_simulation_output


# a global NAME is used for the output directory and for the process name
NAME = 'template'


class Template(Process):
    '''
    This mock process provides a basic template that can be used for a new process
    '''

    # give the process a name, so that it can register in the process_repository
    name = NAME

    # declare default parameters as class variables
    defaults = {
        'uptake_rate': 0.1,
    }

    def __init__(self, parameters=None):
        # parameters passed into the constructor merge with the defaults
        # and can be access through the self.parameters class variable
        super(Template, self).__init__(parameters)

    def ports_schema(self):
        '''
        ports_schema returns a dictionary that declares how each state will behave.
        Each key can be assigned settings for the schema_keys declared in Store:

        * `_default`
        * `_updater`
        * `_divider`
        * `_value`
        * `_properties`
        * `_emit`
        * `_serializer`
        '''

        return {
            'internal': {
                'A': {
                    '_default': 1.0,
                    '_updater': 'accumulate',
                    '_emit': True,
                }
            },
            'external': {
                'A': {
                    '_default': 1.0,
                    '_updater': 'accumulate',
                    '_emit': True,
                }
            },
        }

    def next_update(self, timestep, states):

        # get the states
        internal_a = states['internal']['A']
        external_a = states['external']['A']

        # calculate timestep-dependent updates
        internal_update = self.parameters['uptake_rate'] * external_a * timestep
        external_update = -1 * internal_update

        # return an update that mirrors the ports structure
        return {
            'internal': {
                'A': internal_update},
            'external': {
                'A': external_update}
        }


# functions to configure and run the process
def run_template_process():
    '''Run a simulation of the process.

    Returns:
        The simulation output.
    '''

    # initialize the process by passing in parameters
    parameters = {}
    template_process = Template(parameters)

    # declare the initial state, mirroring the ports structure
    initial_state = {
        'internal': {
            'A': 0.0
        },
        'external': {
            'A': 1.0
        },
    }

    # run the simulation
    composite = template_process.generate()
    sim = Engine(composite=composite, initial_state=initial_state)
    sim.update(10)
    output = sim.emitter.get_timeseries()

    return output


def test_template_process():
    '''Test that the process runs correctly.

    This will be executed by pytest.
    '''
    output = run_template_process()
    # TODO: Add assert statements to ensure correct performance.


def main():
    '''Simulate the process and plot results.'''
    # make an output directory to save plots
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)

    output = run_template_process()

    # plot the simulation output
    plot_settings = {}
    plot_simulation_output(output, plot_settings, out_dir)


# run module with python vivarium/process/template_process.py
if __name__ == '__main__':
    main()



"""
=======================
Swap Processes Process
=======================
"""

import os
from typing import Any, Dict

from vivarium.core.process import (
    Deriver,
)
from vivarium.core.composer import Composer
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.core.engine import Engine
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.composites.toys import ExchangeA
from vivarium.processes.timeline import TimelineProcess

NAME = 'swap_compartment'


class SwapProcesses(Deriver):
    """ SwapProcesses Process

    Replaces the contents of a compartment when the state under the
    'trigger' port is set to True.

    Configuration:

    * **``removed_processes``**: A list of the names of the processes
      that will be removed from the compartment upon trigger.
    * **``new_compartment``**: An instantiated compartment, with
      processes and a topology that will replace the removed_processes
      upon trigger.
    * **``initial_state``**: states that will be set upon trigger.

    :term:`Ports`:

    * **``trigger``**: contains the variable that triggers the swap when True.
    * **``self``**: This port connects to the compartment's path, and
      this is where the `_delete` and `_generate` updates get pointed
      to upon trigger.
    """
    name = NAME
    defaults: Dict[str, Any] = {
        'removed_processes': [],
        'new_compartment': None,
        'initial_state': {},
    }

    def __init__(self, parameters=None):
        super(SwapProcesses, self).__init__(parameters)
        self.removed_processes = self.parameters['removed_processes']
        self.new_compartment = self.parameters['new_compartment']
        self.initial_state = self.parameters['initial_state']

    def ports_schema(self):
        return {
            'trigger': {
                '_default': False,
                '_emit': True},
            'self': {
                '*': {}}}

    def next_update(self, timestep, states):
        if states['trigger']:
            update = {
                'self': {
                    '_delete': self.removed_processes}}
            if self.new_compartment:
                network = self.new_compartment.generate({})  # todo -- pass in config?
                update['self'].update({
                    '_generate': [{
                        'processes': network['processes'],
                        'topology': network['topology'],
                        'initial_state': self.initial_state,
                    }]})
            return update
        else:
            return {}


# test
class ToyDeadCompartment(Composer):
    defaults = {
        'secrete': {
            'secrete_rate': 0.1}}

    def generate_processes(self, config):
        return {
            'secrete': ExchangeA(config['secrete'])}

    def generate_topology(self, config):
        return {
            'secrete': {
                'internal': ('internal',),
                'external': ('external',)}}


class ToyLivingCompartment(Composer):
    defaults = {
        'exchange': {'uptake_rate': 0.1},
        'death': {
            'removed_processes': [
                'exchange',
                'death'],
            'new_compartment': ToyDeadCompartment({})
        }}

    def generate_processes(self, config):
        return {
            'exchange': ExchangeA(config['exchange']),
            'death': SwapProcesses(config['death'])}

    def generate_topology(self, config):
        self_path = ('..', config['agent_id'])
        return {
            'exchange': {
                'internal': ('internal',),
                'external': ('external',)},
            'death': {
                # set the trigger to be the 'dead' state
                'trigger': ('dead',),
                'self': self_path}}


def test_death():
    agent_id = '1'

    # make the composite
    compartment = ToyLivingCompartment({
        'agent_id': agent_id})

    # timeline turns death on
    time_dead = 5
    time_total = 10
    timeline = [
        (0, {('agents', agent_id, 'dead'): False}),
        (time_dead, {('agents', agent_id, 'dead'): True}),
        (time_total, {})]
    timeline_process = TimelineProcess({'timeline': timeline})

    # compose
    composite = compartment.generate(path=('agents', agent_id))
    composite.merge(
        processes={'timeline': timeline_process},
        topology={'timeline': {
            'global': ('global',),
            'agents': ('agents',)}},
    )

    # initial state
    initial_state = {
        'agents': {
            agent_id: {
                'external': {'A': 1},
                'trigger': False}}}

    # simulate
    sim = Engine(
        composite=composite,
        initial_state=initial_state)
    sim.update(time_total)
    output = sim.emitter.get_timeseries()

    # external starts at 1, goes down until death, and then back up
    # internal does the inverse
    external_a = output['agents']['1']['external']['A']
    internal_a = output['agents']['1']['internal']['A']
    assert external_a[0] == 1
    assert external_a[time_dead] < external_a[0]
    assert external_a[time_total] > external_a[time_dead]
    assert internal_a[0] == 0
    assert internal_a[time_dead] > internal_a[0]
    assert internal_a[time_total] < internal_a[time_dead]

    return output


def run_death():
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)
    output = test_death()
    plot_simulation_output(output, {}, out_dir)


if __name__ == '__main__':
    run_death()



"""
===================
Experiment Control
===================

Run experiments and analyses from the command line.
"""

import os
import argparse
import copy

# typing
from typing import (
    Any, Dict, Optional, Union, Sequence, List)
from vivarium.core.types import OutputDict

from vivarium.core.engine import timestamp
from vivarium.core.directories import BASE_OUT_DIR
from vivarium.composites.toys import ToyCompartment, test_composer

from vivarium.plots.simulation_output import plot_simulation_output


def make_dir(out_dir: str = 'out') -> None:
    os.makedirs(out_dir, exist_ok=True)  # pragma: no cover


class Control:
    """ Control experiments from the command line

    Load experiments, plots, and workflows in this Control class,
    and trigger them from the command line
    """

    def __init__(
            self,
            out_dir: Optional[str] = None,
            experiments: Optional[Dict[str, Any]] = None,
            composers: Optional[Dict[str, Any]] = None,
            plots: Optional[Dict[str, Any]] = None,
            workflows: Optional[Dict[str, Any]] = None,
            args: Optional[Sequence[str]] = None,
    ) -> None:
        workflows = workflows or {}
        experiments = experiments or {}
        plots = plots or {}
        workflows = workflows or {}

        self.experiments_library = experiments
        self.compposers_library = composers
        self.plots_library = plots
        self.workflows_library = workflows
        self.output_data = None

        # base output directory
        self.out_dir = BASE_OUT_DIR
        if out_dir:
            self.out_dir = out_dir

        # arguments
        self.args = self.parse_args(args)

        if self.args.experiment:
            experiment_id = str(self.args.experiment)
            self.output_data = self.run_experiment(experiment_id)

        if self.args.workflow:
            workflow_id = str(self.args.workflow)
            self.run_workflow(workflow_id)

    def parse_args(
            self, args: Optional[Sequence[str]] = None
    ) -> argparse.Namespace:
        parser = argparse.ArgumentParser(
            description='command line control of experiments'
        )
        parser.add_argument(
            '--workflow', '-w',
            type=str,
            choices=list(self.workflows_library.keys()),
            help='the workflow id'
        )
        parser.add_argument(
            '--experiment', '-e',
            type=str,
            choices=list(self.experiments_library.keys()),
            help='experiment id to run'
        )
        return parser.parse_args(args)

    def run_experiment(
            self,
            experiment_config: Union[str, dict]
    ) -> OutputDict:

        if isinstance(experiment_config, dict):
            if 'experiment_id' in experiment_config:
                experiment_id = experiment_config.pop('experiment_id')
                experiment = self.experiments_library[experiment_id]
            else:
                experiment = experiment_config.pop('experiment')
            return experiment(**experiment_config)

        if isinstance(experiment_config, str):
            experiment = self.experiments_library[experiment_config]
            if isinstance(experiment, dict):
                experiment = experiment.pop('experiment')
            return experiment()

        raise Exception(f'invalid experiment config: {experiment_config}')

    def run_one_plot(
            self,
            plot_config: Union[str, dict],
            data: OutputDict,
            out_dir: Optional[str] = None,
    ) -> None:
        data_copy = copy.deepcopy(data)

        if isinstance(plot_config, str):
            plot_config = self.plots_library[plot_config]

        if isinstance(plot_config, dict):
            if 'plot_id' in plot_config:
                plot_id = plot_config.pop('plot_id')
                plot = self.plots_library[plot_id]
            else:
                plot = plot_config.pop('plot')
            plot(
                data=data_copy,
                out_dir=out_dir,
                **plot_config)

        elif callable(plot_config):
            # call plot directly
            plot_config(
                data=data_copy,
                out_dir=out_dir)
        else:
            raise Exception(f'invalid plot config: {plot_config}')

    def run_plots(
            self,
            plot_ids: Union[list, str],
            data: OutputDict,
            out_dir: Optional[str] = None,
    ) -> None:

        if isinstance(plot_ids, list):
            for plot_id in plot_ids:
                self.run_one_plot(plot_id, data, out_dir=out_dir)
        else:
            self.run_one_plot(plot_ids, data, out_dir=out_dir)

    def run_workflow(self, workflow_id: str) -> None:
        workflow = self.workflows_library[workflow_id]
        experiment_id = workflow['experiment']
        plot_ids = workflow.get('plots')

        # output directory for this workflow
        workflow_name = workflow.get('name', timestamp())
        out_dir = os.path.join(self.out_dir, workflow_name)

        # run the experiment
        self.output_data = self.run_experiment(experiment_id)

        # run the plots
        if plot_ids:
            self.run_plots(plot_ids, self.output_data, out_dir=out_dir)
            print('plots saved to directory: {}'.format(out_dir))


# testing

def toy_plot(
        data: OutputDict,
        config: Optional[Dict] = None,
        out_dir: Optional[str] = 'out'
) -> None:
    del config  # unused
    plot_simulation_output(data, out_dir=out_dir)


def toy_control(
        args: Optional[Sequence[str]] = None) -> Control:
    """ a toy example of control

    To run:
    > python vivarium/core/control.py -w 1
    """
    experiment_library = {
        # put in dictionary with name
        '1': {
            'name': 'exp_1',
            'experiment': test_composer},
        # map to function to run as is
        '2': test_composer,
    }
    plot_library = {
        # put in dictionary with config
        '1': {
            'plot': toy_plot,
            'config': {}},
        # map to function to run as is
        '2': toy_plot
    }
    composers_library = {
        'agent': ToyCompartment,
    }
    workflow_library = {
        '1': {
            'name': 'test_workflow',
            'experiment': '1',
            'plots': ['1']},
        '2': {
            'name': 'test_workflow',
            'experiment': '1',
            'plots': '2'}
    }

    control = Control(
        out_dir=os.path.join('out', 'control_test'),
        experiments=experiment_library,
        composers=composers_library,
        plots=plot_library,
        workflows=workflow_library,
        args=args,
    )

    return control


def is_float(element: Any) -> bool:
    try:
        float(element)
        return True
    except ValueError:
        return False


def _parse_options(
        options_list: Optional[List[str]]
) -> Dict[str, Union[int, float, bool, str]]:
    """Parse the KEY=VALUE or KEY=k=v option strings into a dict."""
    assignments = options_list or []
    pairs = [
        a.split('=', 1) + [''] for a in assignments
    ]  # [''] to handle the no-'=' case
    options = {}
    for p in pairs:
        key: str = p[0]
        str_value: str = p[1]
        value: Any = None
        if str_value.isdigit():
            value = int(str_value)
        elif is_float(str_value):
            value = float(str_value)
        elif str_value in ['True', 'False']:
            value = bool(str_value)
        else:
            value = str_value
        options[key] = value
    return options


def run_library_cli(library: dict, args: Optional[list] = None) -> None:
    """Run experiments from the command line

    Args:
        library (dict): maps experiment id to experiment function
    """
    parser = argparse.ArgumentParser(
        description='run experiments from the command line')
    parser.add_argument(
        '--name', '-n', default=[], nargs='+',
        help='experiment ids to run')
    parser.add_argument(
        '--options', '-o', metavar='OPTION_KEY=VALUE',
        action='append', help='A "KEY=VALUE" option')
    parser_args = parser.parse_args(args)
    run_all = not parser_args.name
    options = _parse_options(parser_args.options)

    for name in parser_args.name:
        library[name](**options)
    if run_all:
        for name, test in library.items():
            test()


def test_library_cli() -> None:
    def run_fun(key: Any = False) -> dict:
        return {'key': key}
    lib = {'1': run_fun}
    run_library_cli(lib, args=['-n', '1', '-o', 'key=True'])
    run_library_cli(lib, args=['-n', '1', '-o', 'key=0.2'])
    run_library_cli(lib, args=['-n', '1', '-o', 'key=b'])


def test_control() -> None:
    toy_control(args=['-w', '1'])
    toy_control(args=['-w', '2'])
    control = toy_control(args=['-e', '2'])
    control.run_workflow('1')


fun_lib = {
    '0': test_library_cli,
    '1': test_control,
}


# python vivarium/core/control.py -n [test number]
if __name__ == '__main__':
    run_library_cli(fun_lib)



import os

REFERENCE_DATA_DIR = os.path.join('vivarium', 'reference_data')
BASE_OUT_DIR = 'out'
TEST_OUT_DIR = os.path.join(BASE_OUT_DIR, 'tests')
PROCESS_OUT_DIR = os.path.join(BASE_OUT_DIR, 'processes')
COMPARTMENT_OUT_DIR = os.path.join(BASE_OUT_DIR, 'compartments')
COMPOSITE_OUT_DIR = os.path.join(BASE_OUT_DIR, 'composites')
EXPERIMENT_OUT_DIR = os.path.join(BASE_OUT_DIR, 'experiments')



""" Toy Stochastic Transcription Process
Toy model of Gillespie algorithm-based  transcription,
and a composite with deterministic translation.

Note: This Process is primarily for testing multi-timestepping.
variables and parameters are hard-coded. Do not use this as a
general stochastic transcription.
"""
import os
import numpy as np

from vivarium.core.process import Process
from vivarium.core.composer import Composer
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.core.engine import Engine
from vivarium.core.registry import process_registry
from vivarium.plots.simulation_output import plot_simulation_output



class StochasticTSC(Process):
    """stochastic toy transcription"""
    defaults = {
        'ktsc': 5e0,
        'kdeg': 1e-1,
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.ktsc = self.parameters['ktsc']
        self.kdeg = self.parameters['kdeg']
        self.stoichiometry = np.array([[0, 1], [0, -1]])
        self.time_left = None
        self.event = None

        # initialize the next timestep
        initial_state = self.initial_state()
        self.calculate_timestep(initial_state)

    def initial_state(self, config=None):
        return {
            'DNA': {
                'G': 1.0
            },
            'mRNA': {
                'C': 1.0
            }
        }

    def ports_schema(self):
        return {
            'DNA': {
                'G': {
                    '_default': 1.0,
                    '_emit': True}},
            'mRNA': {
                'C': {
                    '_default': 1.0,
                    '_emit': True}}}

    def calculate_timestep(self, states):
        # retrieve the state values
        g = states['DNA']['G']
        c = states['mRNA']['C']

        array_state = np.array([g, c])

        # Calculate propensities
        propensities = [
            self.ktsc * array_state[0], self.kdeg * array_state[1]]
        prop_sum = sum(propensities)

        # The wait time is distributed exponentially
        self.calculated_timestep = np.random.exponential(scale=prop_sum)
        return self.calculated_timestep

    def next_reaction(self, x):
        """get the next reaction and return a new state"""

        propensities = [self.ktsc * x[0], self.kdeg * x[1]]
        prop_sum = sum(propensities)

        # Choose the next reaction
        r_rxn = np.random.uniform()
        i = 0
        for i, _ in enumerate(propensities):
            if r_rxn < propensities[i] / prop_sum:
                # This means propensity i fires
                break
        x += self.stoichiometry[i]
        return x

    def next_update(self, timestep, states):

        if self.time_left is not None:
            if timestep >= self.time_left:
                event = self.event
                self.event = None
                self.time_left = None
                return event

            self.time_left -= timestep
            return {}

        # retrieve the state values, put them in array
        g = states['DNA']['G']
        c = states['mRNA']['C']
        array_state = np.array([g, c])

        # calculate the next reaction
        new_state = self.next_reaction(array_state)

        # get delta mRNA
        c1 = new_state[1]
        d_c = c1 - c

        update = {
            'mRNA': {
                'C': d_c}}

        if self.calculated_timestep > timestep:
            # didn't get all of our time, store the event for later
            self.time_left = self.calculated_timestep - timestep
            self.event = update
            return {}

        # return an update
        return {
            'mRNA': {
                'C': d_c}}


class TRL(Process):
    """deterministic toy translation"""

    defaults = {
        'ktrl': 1e-2,
        'kdeg': 1e-4,
        }

    def ports_schema(self):
        return {
            'mRNA': {
                'C': {
                    '_default': 1.0,
                    '_emit': True}},
            'Protein': {
                'X': {
                    '_default': 1.0,
                    '_emit': True}}}

    def next_update(self, timestep, states):
        c = states['mRNA']['C']
        x = states['Protein']['X']
        d_x = (
            self.parameters['ktrl'] * c -
            self.parameters['kdeg'] * x) * timestep
        return {
            'Protein': {
                'X': d_x}}


class TrlConcentration(TRL):
    """rescale mRNA"""

    def next_update(self, timestep, states):
        states['mRNA']['C'] = states['mRNA']['C'] * 1e5
        return super().next_update(timestep, states)


class StochasticTscTrl(Composer):
    """
    composite toy model with stochastic transcription,
    deterministic translation.
    """
    defaults = {
        'stochastic_TSC': {'time_step': 10},
        'TRL': {'time_step': 10},
    }

    def generate_processes(self, config):
        counts_to_molar = process_registry.access(
            'counts_to_molar')
        return {
            'stochastic_TSC': StochasticTSC(config['stochastic_TSC']),
            'TRL': TrlConcentration(config['TRL']),
            'concs': counts_to_molar({'keys': ['C']})
        }

    def generate_topology(self, config):
        return {
            'stochastic_TSC': {
                'DNA': ('DNA',),
                'mRNA': ('mRNA_counts',)
            },
            'TRL': {
                'mRNA': ('mRNA',),
                'Protein': ('Protein',)
            },
            'concs': {
                'counts': ('mRNA_counts',),
                'concentrations': ('mRNA',)}
        }



def test_gillespie_process(total_time=1000):
    gillespie_process = StochasticTSC({'name': 'process'})

    # make the experiment
    exp_settings = {
        'display_info': False,
        'experiment_id': 'TscTrl'}
    composite = gillespie_process.generate()
    gillespie_experiment = Engine(
        composite=composite,
        **exp_settings)

    # run the experiment in large increments
    increment = 10
    for i in range(total_time):
        if i == total_time - 1:
            gillespie_experiment.run_for(increment, force_complete=True)
            # Now the process is at the global time.
            break
        gillespie_experiment.run_for(increment)
        # check that process remains behind global time
        front = gillespie_experiment.front
        assert front[('process',)]['time'] < gillespie_experiment.global_time

    front = gillespie_experiment.front
    assert front[('process',)]['time'] == gillespie_experiment.global_time

    gillespie_data = gillespie_experiment.emitter.get_timeseries()
    return gillespie_data


def test_gillespie_composite(total_time=10000):
    stochastic_tsc_trl = StochasticTscTrl().generate()

    # make the experiment
    exp_settings = {
        'experiment_id': 'stochastic_tsc_trl'}
    stoch_experiment = Engine(
        composite=stochastic_tsc_trl,
        **exp_settings)

    # simulate and retrieve the data from emitter
    stoch_experiment.update(total_time)
    data = stoch_experiment.emitter.get_timeseries()

    return data


def main():
    """run the tests and plot"""
    out_dir = os.path.join(PROCESS_OUT_DIR, 'toy_gillespie')
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    process_output = test_gillespie_process()
    composite_output = test_gillespie_composite()

    # plot the simulation output
    plot_settings = {}
    plot_simulation_output(
        process_output, plot_settings, out_dir, filename='process')
    plot_simulation_output(
        composite_output, plot_settings, out_dir, filename='composite')


if __name__ == '__main__':
    main()
